#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import datetime
import hashlib
import logging
import os
import sys
import time
from pathlib import Path

from pycapo import CapoConfig

from . import FITSUtil as fi
from . import IngestionManifest as im
from . import NGASUtil as ng
from . import almautils as alu
from . import archiveutils as au
from . import evlasciencedatamodel as sc
from . import persistmetadata as pm
from . import remotecopy as rc
from .IngestionManifest import IngestionManifest
from .JSONMetadataUtil import JSONMetadata
from .NGASUtil import sha256
from .schema import create_session
from .schema.model import (
    AlmaSourceDatum,
    AlmaSpwDatum,
    AncillaryProduct,
    Calibration,
    File,
    Filegroup,
    Image,
    ProductGroup,
    ScienceProduct,
    ScienceProductsProductGroup,
)

config = CapoConfig()
site_name = config.getstring("archive-ingestion.NGASSite")
ngas_cluster = config.getstring("archive-ingestion.NGASCluster")

LOG = logging.getLogger("ingestion")
_SCIENCE_TARGET = "target"

session = create_session("SDM")
ngas_files_to_ingest = []


def ingest_from_manifest(path, manifest):
    LOG.info("Ingestion via ingestion manifest")
    (
        telescope,
        operation,
        reingest,
        ngas_ingest,
        ingestion_path,
        additional_metadata_file,
        collection_metadata_file,
    ) = manifest.get_ingestion_parameters2()
    LOG.debug(f"telescope: {telescope}")
    LOG.debug(f"operation: {operation}")
    LOG.debug(f"reingest: {reingest}")
    LOG.debug(f"ngas_ingest: {ngas_ingest}")
    LOG.debug(f"path: {path}")
    LOG.debug(f"ingestion_path: {ingestion_path}")
    LOG.debug(f"collection metadata file: {collection_metadata_file}")
    input_group = manifest.get_input_group()
    output_group = manifest.get_output_group()
    associate_group = manifest.get_associate_group()
    LOG.debug(f"associaate group: {associate_group}")

    fileset = ""

    if operation == im.NAASC_CALIBRATION:
        mous = manifest.get_output_group_science_products()[0]["filename"]
    elif operation == im.EXECUTION_BLOCK:
        fileset = manifest.get_output_group_science_products()[0]["filename"]
        mous = None
    else:
        mous = None

    # Most of the time: We've got an output_group, so we go through this path
    if output_group is not None:
        input_science_products, output_science_products = persist_science_metadata(
            telescope,
            operation,
            input_group,
            output_group,
            reingest,
            ngas_ingest,
            ingestion_path,
            additional_metadata_file,
            collection_metadata_file,
            path,
        )
        LOG.debug(f"*Input science products-2: {input_science_products}")
        LOG.debug(f"*Output science products-2: {output_science_products}")
        LOG.debug(f"*Output group ancillary products: {output_group['ancillary_products']}")
        # Persist product groups and associated objects (science_product_product_groups, science_product_project)
        persist_groups(
            telescope,
            mous,
            input_science_products,
            output_science_products,
            output_group["ancillary_products"],
            associate_group,
            manifest,
            ingestion_path,
        )
    else:
        # However, adding products after the fact (science or ancillary) are functions too
        # For now, we only have use cases for after-the-fact ancillary products

        ancillary_list = manifest.get_additional_ancillary_products()
        if ancillary_list is not None:

            for product in ancillary_list:
                LOG.debug(product["filename"])

                persist_ancillary_product_metadata(product, ingestion_path, telescope)

                #
                # handle any collection implications
                #
                if collection_metadata_file is not None:
                    collection_data = JSONMetadata(collection_metadata_file)
                    # Only expect ELWA so far:
                    if collection_data.isElwa():
                        #
                        # Simplist case
                        #
                        locator = product["science_associate"]
                        LOG.info("Handling ELWA collection update")
                        science_product = (
                            session.query(ScienceProduct)
                            .filter(ScienceProduct.science_product_locator == locator)
                            .one()
                        )
                        science_product.collection = "ELWA"
                        session.add(science_product)
            # Then, commit the list here.
            # in practice, it's just the one file/call for ELWA right now
            session.commit()
        else:
            LOG.error("Neither output_group nor ancillary_products list defined, cannot continue!")
            sys.exit(-5)

    if ngas_ingest and len(ngas_files_to_ingest) > 0:
        ngas_archive(ngas_files_to_ingest, ingestion_path, fileset)


def persist_science_metadata(
    telescope,
    operation,
    input_group,
    output_group,
    reingest,
    ngas_ingest,
    ingestion_path,
    additional_metadata_file,
    collection_metadata_file,
    path,
):
    if input_group is not None:
        # Input group will be empty for SDM ingestion
        input_group_science_products = input_group["science_products"]
    else:
        input_group_science_products = []
        input_science_products = []
    output_group_science_products = output_group["science_products"]
    output_group_ancillary_products = output_group["ancillary_products"]

    LOG.debug(f"Operation: {operation}")

    if operation == im.CALIBRATION:
        input_science_products, output_science_products = persist_calibration_metadata(
            telescope,
            input_group_science_products,
            output_group_science_products,
            output_group_ancillary_products,
            reingest,
            ngas_ingest,
            ingestion_path,
        )
    elif operation == im.NAASC_CALIBRATION:
        input_science_products, output_science_products = persist_naasc_calibration_metadata(
            telescope,
            input_group_science_products,
            output_group_science_products,
            output_group_ancillary_products,
            reingest,
            ngas_ingest,
            ingestion_path,
        )
    elif operation == im.EXECUTION_BLOCK:
        output_science_products = persist_sdm(
            telescope,
            output_group_science_products,
            reingest,
            ingestion_path,
            additional_metadata_file,
            collection_metadata_file,
            path,
        )
    elif operation == im.CONTINUUM_IMAGE or im.IMAGE_CUBE:
        input_science_products, output_science_products = persist_images(
            telescope,
            input_group_science_products,
            output_group_science_products,
            output_group_ancillary_products,
            reingest,
            ingestion_path,
            additional_metadata_file,
            collection_metadata_file,
            path,
        )
    else:
        LOG.critical(f"{operation} is not a valid operation")
        sys.exit(1)

    return input_science_products, output_science_products


def persist_sdm(
    telescope,
    output_group_science_products,
    reingest,
    ingestion_path,
    additional_metadata_file,
    collection_metadata_file,
    path,
):
    LOG.debug(f"SDM OGSP: {output_group_science_products}")
    input_science_products = []
    output_science_products = []
    output_ancillary_products = []

    sdm_identifier = output_group_science_products[0]["filename"]
    if telescope == "ALMA":
        # ASDM UIDs are not disk-friendly, so look for a modified name:
        sdm_identifier = sdm_identifier.replace(":", "_").replace("/", "_")

    exec_block_file = Path(ingestion_path) / sdm_identifier
    if not exec_block_file.exists():
        LOG.error(f"File set {exec_block_file.name} not found in {ingestion_path}.")
        sys.exit(1)
    # Now make sure the ASDM file exists
    asdm_file = exec_block_file / "ASDM.xml"
    if not asdm_file.exists():
        LOG.error(f"ASDM.xml not found in {exec_block_file.name}")
        sys.exit(1)

    execution_block = au.getexecutionblockfromname(exec_block_file.name)

    if execution_block is not None:
        if reingest:
            LOG.debug(f"Re-ingestion requested")
        else:
            LOG.error(
                f"Execution block {execution_block.ngas_fileset_id} already exists in archive and re-ingestion not requested."
            )
            sys.exit(1)

    else:
        LOG.debug(f"Initial ingestion of {exec_block_file.name}")

        sdm = sc.buildsdm(exec_block_file.name, ingestion_path)
        if len(sdm.bdffiles) > 0:  # ingest science meta data
            LOG.debug(f"SDM for {sdm.fileset} has {len(sdm.bdffiles)} BDFs")
            LOG.debug(f"SDM Tables {sdm.sdmtables.keys()}")
            starttime = time.time()
            additional_metadata = None
            if additional_metadata_file is not None:
                additional_metadata_file = Path(path) / additional_metadata_file
                if additional_metadata_file.exists():
                    # Execution block has additional metadata
                    additional_metadata = JSONMetadata(additional_metadata_file)
            collection_metadata = None
            if collection_metadata_file is not None:
                collection_metadata_file = Path(path) / collection_metadata_file
                if collection_metadata_file.exists():
                    # Execution block has additional metadata
                    collection_metadata = JSONMetadata(collection_metadata_file)
            project, science_products, ingest_files = pm.persist(
                sdm,
                False,
                additional_metadata,
                collection_metadata,
                output_group_science_products,
                ingestion_path,
                ngas_cluster,
            )
            ngas_files_to_ingest.extend(ingest_files)
            LOG.info(f"SDM metadata ingestion complete for {sdm.fileset} in {round(time.time() - starttime, 2)}s")
        else:
            LOG.error(f"Execution block: {sdm.fileset} has no bdfs and will not be ingested")

    return science_products


def persist_calibration_metadata(
    telescope,
    input_group_science_products,
    output_group_science_products,
    output_group_ancillary_products,
    reingest,
    ngas_ingest,
    ingestion_path,
):
    LOG.debug(f"Persisting {telescope} calibration")
    if not valid_calibration_groups(
        telescope, input_group_science_products, output_group_science_products, output_group_ancillary_products
    ):
        sys.exit(1)

    LOG.debug(f"IG SP: {input_group_science_products}")
    LOG.debug(f"OG SP: {output_group_science_products}")
    LOG.debug(f"OG AP: {output_group_ancillary_products}")

    input_science_products = []
    output_science_products = []

    for science_product in input_group_science_products:
        input_science_products.append(au.get_science_product_by_spl(science_product["locator"]))

    for science_product in output_group_science_products:
        calibration_filename = science_product["filename"]
        tar_file = Path(ingestion_path) / calibration_filename
        if not tar_file.exists():
            LOG.warning(f"Calibration products file: {tar_file.name} does not exist in {tar_file.parent}.")
            sys.exit(1)
        calibration_tar_file = au.get_file(calibration_filename)
        if calibration_tar_file is not None:  # The file already exists
            if reingest:
                LOG.info("Re-ingestion")
                calibration_science_product = au.get_science_product(calibration_filename)
                calibration_science_product.metadata_ingestion_date = datetime.datetime.utcnow()
                calibration_science_product.metadata_ingestion_version = (
                    int(calibration_science_product.metadata_ingestion_version) + 1
                )
                output_science_products.append(calibration_science_product)
                LOG.debug(f"Output science products: {output_science_products}")

                calibration_tar_file.ingestion_time = datetime.datetime.utcnow()
                calibration_tar_file.checksum = sha256(Path(ingestion_path) / calibration_tar_file.filename)
                db_session = session.object_session(calibration_science_product)
                db_session.add(calibration_science_product)
                db_session.add(calibration_tar_file)
                ngas_files_to_ingest.append((calibration_tar_file, calibration_tar_file.filesize))
            else:
                LOG.error(f"The file {calibration_filename} exists and re-ingestion not requested. Exiting.")
                sys.exit(1)
        else:
            LOG.debug("Initial ingestion")

            calibration_science_product_locator = ng.generate_science_product_locator(
                input_science_products[0].execution_block.telescope.lower(), science_product["type"]
            )

            # Create calibration Filegroup and File
            calibration_filegroup = Filegroup(datasize=tar_file.stat().st_size, type="calibration")

            calibration_products_file = File(
                file_path=str(tar_file.parent),
                ngas_id=ng.generate_uuids_for_ngas(telescope, "calibration", "tar"),
                ngas_location=site_name,
                ngas_cluster=ngas_cluster,
                filename=tar_file.name,
                filegroup1=calibration_filegroup,
                filesize=calibration_filegroup.datasize,
                format="tar",
                type="cal",
                checksum=sha256(tar_file),
                checksum_type="SHA256",
                ingestion_time=datetime.datetime.utcnow(),
            )

            ngas_files_to_ingest.append((calibration_products_file, calibration_products_file.filesize))

            calibration = Calibration(science_product_locator=calibration_science_product_locator)

            calibration.execution_block = input_science_products[0].execution_block
            calibration.project = input_science_products[0].execution_block.project
            calibration.filegroup = calibration_filegroup

            # Create ScienceProduct
            output_science_product = ScienceProduct(
                science_product_locator=calibration_science_product_locator,
                science_product_type="Calibration",
                metadata_ingestion_date=datetime.datetime.utcnow(),
                filegroup=calibration_filegroup,
                metadata_ingestion_version=1,
                external_system="EVLA Processing",
                external_name=tar_file.name,
            )

            output_science_products.append(output_science_product)

            # Update science_products_projects
            calibration.execution_block.project.science_products.append(output_science_product)

            db_session = session.object_session(calibration.execution_block)
            db_session.add(calibration_filegroup)
            db_session.add(calibration_products_file)
            db_session.add(calibration)
            db_session.add(output_science_product)

    LOG.debug(f"Input science products: {input_science_products}")
    LOG.debug(f"Output science products: {output_science_products}")

    db_session.commit()

    # output_auxiliary_products will only contain auxiliary products associated with science products
    # auxiliary products associated with a group will be added when groups are created

    return input_science_products, output_science_products


def persist_naasc_calibration_metadata(
    telescope,
    input_group_science_products,
    output_group_science_products,
    output_group_ancillary_products,
    reingest,
    ngas_ingest,
    ingestion_path,
):
    LOG.debug(f"Persisting {telescope} calibration")
    if not valid_calibration_groups(
        telescope, input_group_science_products, output_group_science_products, output_group_ancillary_products
    ):
        sys.exit(1)

    LOG.debug(f"IG SP: {input_group_science_products}")
    LOG.debug(f"OG SP: {output_group_science_products}")
    LOG.debug(f"OG SP: {output_group_ancillary_products}")

    input_science_products = []
    output_science_products = []
    eb_locator_list = []

    for science_product in input_group_science_products:
        input_science_products.append(au.get_science_product_by_spl(science_product["locator"]))
        eb_locator_list.append(science_product["locator"])

    # Returns AlmaOus object
    mous = au.get_mous_from_eb_science_product_locator_list(eb_locator_list)
    session.add(mous)

    LOG.debug(f"Input science products: {input_group_science_products}")

    science_product_locator = ng.generate_science_product_locator(telescope.lower(), "calibration")
    session.add(input_science_products[0])
    project_code = input_science_products[0].filegroup.project_code

    LOG.debug(f"Project code: {project_code}")

    calibration_filegroup = Filegroup(project_code=project_code, datasize=0, type="calibration")

    LOG.debug(
        f"Calibration filegroup: project_code={calibration_filegroup.project_code} type={calibration_filegroup.type}"
    )

    calibration_files = alu.getalmacalfiles(mous.alma_ous_id)

    for calibration_file in calibration_files:
        file = File(
            ngas_id=calibration_file[0],
            ngas_location="NAASC",
            ngas_cluster="NAASC",
            filename=calibration_file[0],
            filesize=calibration_file[1],
            filegroup1=calibration_filegroup,
            format="cal",
            type="cal",
            ingestion_time=datetime.datetime.utcnow(),
        )
        calibration_filegroup.datasize += file.filesize
        LOG.debug(
            f"File ngas_id={file.ngas_id} ngas_location={file.ngas_location} filename={file.filename} filesize={file.filesize} ingestion time: {file.ingestion_time}"
        )
        session.add(file)
    session.add(calibration_filegroup)

    calibration = Calibration(
        science_product_locator=science_product_locator,
        alma_ous=mous,
        project=input_science_products[0].filegroup.project,
        filegroup=calibration_filegroup,
    )

    LOG.debug(
        f"Calibration science product locator={calibration.science_product_locator} alma ous={calibration.alma_ous.alma_ous_id} project code={calibration.project.project_code}"
    )

    session.add(calibration)

    output_science_product = ScienceProduct(
        science_product_locator=science_product_locator,
        science_product_type="Calibration",
        metadata_ingestion_date=datetime.datetime.utcnow(),
        filegroup=calibration_filegroup,
        metadata_ingestion_version=1,
        external_system="ALMA Processing",
        external_name=mous.alma_ous_id,
    )

    output_science_products.append(output_science_product)
    session.add(output_science_product)

    # Update science_products_projects - this is the problem. No, this is working now
    calibration.project.science_products.append(output_science_product)

    LOG.debug(
        f"Science product science product locator={output_science_product.science_product_locator} science product type={output_science_product.science_product_type} metadata ingestion date={output_science_product.metadata_ingestion_date} metadata ingestion version={output_science_product.metadata_ingestion_version}"
    )

    source_list = alu.get_source_list(mous.alma_ous_id)

    for source in source_list:
        alma_source = AlmaSourceDatum(
            alma_ous_id=mous.alma_ous_id,
            field_name=source.source_name,
            ra=source.ra_source,
            dec=source.dec_source,
            integration_time=source.int_time,
            angular_resolution=source.spatial_scale_min,
            largest_angular_scale=source.spatial_scale_max,
            science_field=_SCIENCE_TARGET in source.scan_intent.lower(),
        )

        LOG.debug(
            f"AlmaSourceDatum alma_ouss_id={alma_source.alma_ous_id} field name={alma_source.field_name} ra={alma_source.ra} dec={alma_source.dec} integration time={alma_source.integration_time}"
        )

        spectral_window_data = alu.get_spectral_window_data(source)
        for spw in spectral_window_data:
            spectral_window = AlmaSpwDatum(
                alma_ous_id=mous.alma_ous_id,
                spw_name=spw.spectral_window_name,
                min_frequency=spw.frequency_min,
                max_frequency=spw.frequency_max,
                bandwidth=spw.bandwidth,
                num_channels=spw.channel_num,
                spectral_resolution=spw.resolution_max,
            )
            if spw.resolution_min != spw.resolution_max:
                spectral_window.spectral_resolution = (spw.resolution_max + spw.resolution_min) / 2

            LOG.debug(f"AlmaSpwDatum alma_ous id={spectral_window.alma_ous_id} spw name: {spectral_window.spw_name}")

            session.merge(spectral_window)
            alma_source.spectral_windows.append(spectral_window)
        session.merge(alma_source)

    session.commit()

    return input_science_products, output_science_products


def persist_images(
    telescope,
    input_group_science_products,
    output_group_science_products,
    output_group_ancillary_products,
    reingest,
    ingestion_path,
    metadata_file,
    collection_metadata_file,
    path,
):
    """
    Persist an image (quicklook, image cube, etc.)

    :param telescope:
    :param input_group_science_products:
    :param output_group_science_products:
    :param output_group_ancillary_products:
    :param reingest:
    :param ngas_ingest:
    :param ingestion_path:
    :param metadata_file:
    :param path:
    :return:
    """

    LOG.debug(f"IG SP: {input_group_science_products}")
    LOG.debug(f"OG SP: {output_group_science_products}")
    LOG.debug(f"OG AP: {output_group_ancillary_products}")

    input_science_products = []
    output_science_products = []

    # JSON metadata
    json_metadata_file = Path(path) / metadata_file
    if not json_metadata_file.exists():
        LOG.critical(f"Metadata file {json_metadata_file.name} does not exist in {json_metadata_file.parent}")
        sys.exit(1)
    json_metadata = JSONMetadata(json_metadata_file)
    json_metadata.from_json(json_metadata_file)
    project_code = json_metadata.get_project_code()

    project = au.getproject(project_code)

    LOG.debug(f"Project code: {project.project_code}")
    if project is None:
        LOG.error(f"Project {project.project_code} not found")
        sys.exit(1)

    for science_product in input_group_science_products:
        input_science_products.append(au.get_science_product_by_spl(science_product["locator"]))

    for science_product in output_group_science_products:
        science_product_type = science_product["type"]
        LOG.debug(f"Science product type: {science_product_type}")
        image_file = Path(ingestion_path) / science_product["filename"]
        LOG.debug(f"Image path: {image_file.parent} Image file: {image_file.name}")
        science_product_locator = ng.generate_science_product_locator(telescope.lower(), science_product_type)

        # Create image Filegroup and File
        filegroup = Filegroup(
            parent_filegroup=None,
            # parent_filegroup=input_science_products[0].execution_block.filegroup,
            type=science_product_type,
        )

        file = File(
            file_path=str(image_file.parent),
            ngas_id=ng.generate_uuids_for_ngas(telescope, science_product_type, image_file.suffix[1:]),
            ngas_location=site_name,
            ngas_cluster=ngas_cluster,
            filename=image_file.name,
            filegroup1=filegroup,
            filesize=image_file.stat().st_size,
            format="fits",
            type=science_product_type,
            checksum=sha256(image_file),
            checksum_type="SHA256",
            ingestion_time=datetime.datetime.utcnow(),
        )

        ngas_files_to_ingest.append((file, file.filesize))

        session.add(filegroup)
        session.add(file)

        (
            source_name,
            telescope,
            spatial_resolution,
            field_of_view,
            min_intensity,
            max_intensity,
            ra,
            dec,
            min_freq,
            max_freq,
            rest_frequency,
            ra_element_count,
            dec_element_count,
            image_units,
            beam_axis_ratio,
            polarization_id,
            rms_noise,
            spatial_region,
            ra_pixel_size,
            dec_pixel_size,
        ) = fi.get_fits_data(image_file)

        start_time = json_metadata.get_starttime()
        if start_time is None:
            start_time = 0
        end_time = json_metadata.get_endtime()
        if end_time is None:
            end_time = 0
        if rms_noise is None and json_metadata.get_rms_noise() is None:
            rms_noise = 0

        image = Image(
            target_name=source_name,
            telescope=telescope,
            spatial_resolution=spatial_resolution,
            image_field_of_view=field_of_view,
            max_intensity=max_intensity,
            min_intensity=min_intensity,
            rms_noise=rms_noise,
            polarization_id=polarization_id,
            ra=ra,
            dec=dec,
            min_frequency=min_freq,
            max_frequency=max_freq,
            ra_element_count=ra_element_count,
            dec_element_count=dec_element_count,
            starttime=start_time,
            endtime=end_time,
            exposure_time=end_time,
            rest_frequency=rest_frequency,
            image_units=image_units,
            spatial_region=spatial_region,
            beam_axis_ratio=beam_axis_ratio,
            band_code=json_metadata.get_band_code(),
            ra_pixel_size=ra_pixel_size,
            dec_pixel_size=dec_pixel_size,
            tags=json_metadata.get_image_tags(),
            collection_name=json_metadata.get_collection_name(),
            calibration_level=json_metadata.get_calibration_level(),
            science_product_locator=science_product_locator,
            file=file,
        )

        # Ignore re-ingestion for now, but how would you do this

        external_system = telescope + " Processing"
        output_science_product = ScienceProduct(
            science_product_locator=science_product_locator,
            science_product_type="Image",
            metadata_ingestion_date=datetime.datetime.utcnow(),
            filegroup=filegroup,
            metadata_ingestion_version=1,
            external_system=external_system,
            external_name=image_file.name,
        )

        session.add(image)
        session.add(output_science_product)

        output_science_product.project = project
        session.add(output_science_product)

        output_science_products.append(output_science_product)

        # If this science product has ancillary products
        if "ancillary_products" in science_product.keys():
            for anc_product in science_product["ancillary_products"]:
                LOG.debug(f"Ancillary product: {anc_product}")
    product_file = Path(ingestion_path) / anc_product["filename"]
    ancillary_product_locator = ng.generate_science_product_locator(telescope.lower(), "ancillary_product")
    product_extension = product_file.suffix[1:]

    ancillary_product_filegroup = Filegroup(
        type="ancillary_product",
        datasize=-1,
    )
    if anc_product["type"] == "thumbnail_image":
        thumbnail_path = str(Path(ingestion_path) / product_file.name)
        preview_path = str(Path(calculate_thumbnail_path(thumbnail_path)[0]).parent)
    else:
        preview_path = None
    ancillary_product_file = File(
        ngas_id=ng.generate_uuids_for_ngas(telescope, "ancillary_product", product_extension),
        ngas_location=site_name,
        ngas_cluster=ngas_cluster,
        filename=product_file.name,
        filesize=product_file.stat().st_size,
        format=product_extension,
        type=anc_product["type"],
        checksum=sha256(product_file),
        checksum_type="SHA256",
        ingestion_time=datetime.datetime.utcnow(),
        filegroup1=ancillary_product_filegroup,
        preview_storage_path=preview_path,
    )
    ngas_files_to_ingest.append((ancillary_product_file, ancillary_product_file.filesize))
    ancillary_product = AncillaryProduct(
        ancillary_product_locator=ancillary_product_locator,
        ancillary_product_type=anc_product["type"],
        filegroup=ancillary_product_filegroup,
        science_product_locator=science_product_locator,
    )
    session.add(ancillary_product_file)
    session.add(ancillary_product)

    session.commit()

    return input_science_products, output_science_products


#
# Stolen from the process for images, and mildly reworked for the ELWA case
#
def persist_ancillary_product_metadata(anc_product, ingestion_path, telescope):
    LOG.debug(f"Ancillary product: {anc_product}")
    product_file = Path(ingestion_path) / anc_product["filename"]
    ancillary_product_locator = ng.generate_science_product_locator(telescope.lower(), "ancillary_product")
    product_extension = product_file.suffix[1:]
    # Handle weird ELWA extensions (FITS_N)
    if product_extension.lower().find("fits") != -1:
        product_extension = "fits"

    ancillary_product_filegroup = Filegroup(
        type="ancillary_product",
        datasize=-1,
    )
    if anc_product["type"] == "thumbnail_image":
        thumbnail_path = str(Path(ingestion_path) / product_file.name)
        preview_path = str(Path(calculate_thumbnail_path(thumbnail_path)[0]).parent)
    else:
        preview_path = None
    ancillary_product_file = File(
        ngas_id=ng.generate_uuids_for_ngas(telescope, "ancillary_product", product_extension),
        ngas_location=site_name,
        ngas_cluster=ngas_cluster,
        filename=product_file.name,
        filesize=product_file.stat().st_size,
        format=product_extension,
        type=anc_product["type"],
        checksum=sha256(product_file),
        checksum_type="SHA256",
        ingestion_time=datetime.datetime.utcnow(),
        filegroup1=ancillary_product_filegroup,
        preview_storage_path=preview_path,
    )
    ngas_files_to_ingest.append((ancillary_product_file, ancillary_product_file.filesize))

    #
    # Direct, or group ancillary?
    #
    if "science_associate" in anc_product:
        ancillary_product = AncillaryProduct(
            ancillary_product_locator=ancillary_product_locator,
            ancillary_product_type=anc_product["type"],
            filegroup=ancillary_product_filegroup,
            science_product_locator=anc_product["science_associate"],
        )
    elif "group_with" in anc_product:
        #
        # Need to lookup the group to which this product belongs:
        #   Use case not needed yet (as of Nov 2021)
        product_group = -1

        ancillary_product = AncillaryProduct(
            ancillary_product_locator=ancillary_product_locator,
            ancillary_product_type=anc_product["type"],
            filegroup=ancillary_product_filegroup,
            product_group_id=product_group,
        )
    else:
        LOG.error("Incomplete data provided with after-the-fact ancillary product")
        exit(-1)

    session.add(ancillary_product_file)
    session.add(ancillary_product)


def persist_groups(
    telescope,
    mous,
    input_science_products,
    output_science_products,
    output_group_ancillary_products,
    associate_group,
    manifest,
    ingestion_path,
):
    LOG.debug(f"ISPs: {input_science_products}")
    LOG.debug(f"OSPs: {output_science_products}")
    LOG.debug(f"OGAPs: {output_group_ancillary_products}")
    LOG.debug(f"mous: {mous}")

    input_product_group = ProductGroup(product_group_type="pipeline input")

    db_session = session.object_session(output_science_products[0])

    input_product_group.science_products.extend(input_science_products)

    output_product_group = ProductGroup(
        product_group_type="pipeline output", parent_product_group=input_product_group, alma_ous_id=mous
    )

    output_product_group.science_products.extend(output_science_products)

    for anc_product in output_group_ancillary_products:
        product_file = Path(ingestion_path) / anc_product["filename"]
        ancillary_product_locator = ng.generate_science_product_locator(telescope.lower(), "ancillary_product")
        ancillary_product_filegroup = Filegroup(
            type="ancillary_product",
            datasize=-1,
        )

        if anc_product["type"] == "thumbnail_image":
            thumbnail_path = str(Path(ingestion_path) / product_file.name)
            preview_path = str(Path(calculate_thumbnail_path(thumbnail_path)[0]).parent)
        else:
            preview_path = None

        ancillary_product_file = File(
            ngas_id=ng.generate_uuids_for_ngas(telescope, anc_product["type"], product_file.suffix[1:]),
            ngas_location=site_name,
            ngas_cluster=ngas_cluster,
            filename=product_file.name,
            filesize=product_file.stat().st_size,
            format=product_file.suffix[1:],
            type=anc_product["type"],
            checksum=sha256(product_file),
            checksum_type="SHA256",
            ingestion_time=datetime.datetime.utcnow(),
            filegroup1=ancillary_product_filegroup,
            preview_storage_path=preview_path,
        )

        ngas_files_to_ingest.append((ancillary_product_file, ancillary_product_file.filesize))

        ancillary_product = AncillaryProduct(
            ancillary_product_locator=ancillary_product_locator,
            ancillary_product_type=anc_product["type"],
            filegroup=ancillary_product_filegroup,
            product_group=output_product_group,
        )
        db_session.add(ancillary_product_file)
        db_session.add(ancillary_product)

    db_session.add(input_product_group)
    db_session.add(output_product_group)

    # Associate group links ebs
    if associate_group is not None:
        donor_science_product = au.get_science_product_by_spl(manifest.get_associate_group_locator())
        if donor_science_product is None:
            LOG.critical(f"Science product for donor eb not found: {manifest.get_associate_group_locator()}")
        else:
            associate_product_group = ProductGroup(
                product_group_type="associate",
            )
            # Add the SPL for the associated eb
            associate_product_group.science_products.extend(output_science_products)
            # Add the SPL for the donor eb
            local_donor_science_product = db_session.merge(donor_science_product)
            associate_product_group.science_products.extend([local_donor_science_product])

            db_session.add(associate_product_group)

    db_session.commit()


def valid_calibration_groups(telescope, input_science_products, output_science_products, output_ancillary_products):
    if telescope.lower() == "evla":
        if len(input_science_products) != 1:
            print(f"{len(input_science_products)} science products in input group. Only 1 execution block is permitted")
            return False
        if len(output_science_products) != 1:
            print(
                f"{len(output_science_products)} science products in output group. Only 1 calibration product is permitted"
            )
            return False
    elif telescope.lower() == "alma":
        return True
        if len(output_science_products) != 1:
            print(
                f"{len(output_science_products)} science products in output group. Only 1 calibration product is permitted"
            )
            return False
    else:
        print(f"Invalid telescope in manifest: {telescope}")
        return False
    return True


def ngas_archive(files_to_archive, ingestion_path, fileset=""):
    LOG.debug(f"Ingesting {len(files_to_archive)} files into ngas....")
    thumbnails = []
    files_to_ingest = []
    for file in files_to_archive:
        # Some convenience variables:
        file_type = file[0].type
        file_format = file[0].format
        if file_type == "raw":
            if file_format == "sdm":
                path_to_file = Path(ingestion_path) / fileset / file[0].filename
                path_to_ngas_rename = Path(ingestion_path) / fileset / file[0].ngas_id
            elif file_format == "bdf":
                path_to_file = Path(ingestion_path) / fileset / "ASDMBinary" / file[0].filename
                path_to_ngas_rename = Path(ingestion_path) / fileset / "ASDMBinary" / file[0].ngas_id
        else:
            path_to_file = Path(ingestion_path) / file[0].filename
            path_to_ngas_rename = Path(ingestion_path) / file[0].ngas_id

        if file_type == "thumbnail_image":
            # Thumbnails need to go to the cache as well as NGAS,
            # track those for the additional step.
            thumbnails.append(path_to_file)

        if path_to_file != path_to_ngas_rename:
            if os.access(path_to_ngas_rename, os.F_OK):
                # If the file's NGAS name exists in the staging area already, it means this file was
                # recently ingested, and we're re-ingesting an updated version.  Remove the
                # old hard link to facilitate the new data.
                os.unlink(path_to_ngas_rename)
            # There are cases, in particular with old calibrations, where the
            # ngas id is the filename.  That causes os.link to nuke the file
            # which is just bad news. So, don't link if we don't have to.
            os.link(path_to_file, path_to_ngas_rename)

        # add the file with appropriate NGAS naming to the to-do list:
        # LOG.debug(f"Adding {path_to_ngas_rename} to the ingestion list")
        files_to_ingest.append([str(path_to_ngas_rename), file[1]])

    # With everything prepped, make the NGAS calls.
    # LOG.debug(f"Total set of files: {files_to_ingest}")
    ng.archivefiles(files_to_ingest)

    # Handle the thumbnails:
    LOG.debug(f"Thumbs: {thumbnails}")
    # copy_thumbnails(thumbnails, ingestion_path)


def calculate_thumbnail_path(thumbnail_image_file):
    sha1 = hashlib.sha1()

    with open(str(thumbnail_image_file), "rb") as f:
        for block in iter(lambda: f.read(1024), b""):
            sha1.update(block)
    key = sha1.hexdigest()

    # $1 is the first two characters of the sha1 sum, $2 the second two, and $3 the third two
    # ea887dea73301fab8b9da760014449b59e1211ae
    # $ROOT/ea/88/7d/thumbnail

    root_thumbnail_path = config.getstring("archive-ingestion.vlass.ThumbnailPath")

    # Return path without and with ROOT
    return key[0:2] + "/" + key[2:4] + "/" + key[4:6] + "/" + os.path.basename(
        thumbnail_image_file
    ), root_thumbnail_path + "/" + key[0:2] + "/" + key[2:4] + "/" + key[4:6] + "/" + os.path.basename(
        thumbnail_image_file
    )


# def copy_thumbnail(thumbnail_image_file, thumbnail_destination):
#     if not os.path.isdir(os.path.dirname(thumbnail_destination)):
#         os.makedirs(os.path.dirname(thumbnail_destination))
#     shutil.copyfile(thumbnail_image_file, thumbnail_destination)


def copy_thumbnails(thumbnails, ingestion_path):
    for thumbnail in thumbnails:
        # full path to the png:
        thumb_source = Path(ingestion_path) / thumbnail
        # figure out where it goes (based on checksum):
        doubled_locations = calculate_thumbnail_path(thumb_source)
        #
        # The above returns the resultant path with and without the root value from CAPO
        #   we want the second value of the returned tuple
        #
        rc.copy_thumbnail(thumb_source, doubled_locations[1])


def filesdump(ingest_files):
    for file in ingest_files:
        LOG.info(f"File name: {file[0].filename} Size: {file[1]}")


def test1():
    path = Path(".")
    ingest_from_manifest(path.name, im.INGESTION_MANIFEST_FILE)


if __name__ == "__main__":
    test1()
