#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

""" Make sure we've got all the capability routes"""

# pylint: disable=E0401, R0915, W0621
import copy
import http
from datetime import datetime
from typing import List
from unittest.mock import MagicMock, patch

import pytest
from capability.views.capability_execution import (
    create_version_from_previous_execution_script,
)
from capability.views.capability_request import (
    assign_capability_request,
    create_follow_on_capability_request,
)
from pyramid.config import Configurator
from pyramid.interfaces import IRoutesMapper
from pyramid.testing import DummyRequest

from workspaces.capability.enums import CapabilityRequestState
from workspaces.capability.schema import (
    Capability,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityVersion,
)
from workspaces.workflow.services.workflow_info import WorkflowInfo
from workspaces.workflow.services.workflow_service import WorkflowService

RouteList = List[str]


@pytest.fixture()
def capability_routes() -> RouteList:
    """
    All the capability routes (or should be)

    :return: list of all available capability routes
    """
    return [
        "add_file_to_latest",
        "assign_capability_request",
        "cancel_capability_request",
        "change_srdp_compatibility_version_metadata",
        "create_and_submit_capability_request",
        "create_capability",
        "create_capability_request",
        "create_capability_version",
        "create_follow_on_capability_request",
        "create_version_from_previous_execution_script",
        "delete_capability_request",
        "disable_capability",
        "edit_capability",
        "edit_capability_request",
        "edit_file",
        "edit_version_metadata",
        "enable_capability",
        "get_execution_by_version",
        "get_execution_from_workflow_id",
        "home",
        "pause_capability",
        "retrieve_available_qa_staff_in_group",
        "submit_capability_version",
        "cancel_capability_version",
        "unpause_capability",
        "view_active_capability_requests",
        "view_all_versions",
        "view_capability",
        "view_capability_request",
        "view_created_capability_requests",
        "view_enabled_capabilities",
        "view_latest_version",
        "view_specific_version",
        "view_version_metadata",
    ]


def test_routes_exist(test_config: Configurator, capability_routes: RouteList):
    """
    Tests that the Pyramid config has the proper capability routes set up

    :param test_config: Mock pyramid config for testing purposes
    """

    test_config.include("capability.routes")
    mapper = test_config.registry.queryUtility(IRoutesMapper)

    route_names = [route.name for route in mapper.get_routes()]

    for route in route_names:
        assert route in capability_routes

    for route in capability_routes:
        assert route in route_names

    assert len(capability_routes) == len(route_names)


def test_create_follow_on_capability_request(
    request_null_capability: DummyRequest,
):
    """
    Test for capability_request.create_follow_on_capability_request()

    :param request_null_capability: plays a CapabilityRequest on TV

    :return:
    """

    original_matchdict = request_null_capability.matchdict

    fake_capability = Capability()
    fake_capability.name = "null"
    fake_capability_request = CapabilityRequest()
    fake_capability_request.id = 1
    fake_capability_request.capability = fake_capability
    fake_capability_request.capability_name = fake_capability.name

    fake_v1 = CapabilityVersion()
    fake_v1.version_number = 1
    fake_v1.capability = fake_capability
    fake_v1.capability_name = fake_capability.name
    fake_v1.capability_request_id = fake_capability_request.id
    fake_v1.request = fake_capability_request

    fake_v2 = copy.copy(fake_v1)
    fake_v2.version_number = 2
    fake_v2.parameters = {"user_email": "nobody@nowhere.edu"}

    fake_exec = CapabilityExecution()
    fake_exec.created_at = fake_exec.updated_at = datetime.now()
    fake_exec.capability = fake_capability
    fake_exec.capability_request_id = fake_capability_request.id
    fake_exec.request = fake_capability_request
    fake_exec.capability_name = fake_capability.name
    fake_exec.version_number = 2
    fake_exec.state_name = "Complete"
    fake_exec.version = fake_v2
    fake_v2.executions = [fake_exec]

    workflow_message_handler = MagicMock()
    workflow_message_handler.archive_router = MagicMock()
    workflow_message_handler.message_router = MagicMock()
    workflow_reacquire = MagicMock()

    fake_wf_info = WorkflowInfo(workflow_message_handler)
    fake_wf_info.session = workflow_message_handler
    fake_wf_service = WorkflowService(fake_wf_info, workflow_message_handler, workflow_reacquire)

    # set up the *Pyramid* request
    try:
        request_null_capability.matchdict["id"] = 1
        request_null_capability.matchdict["request_id"] = fake_capability_request.id
        request_null_capability.matchdict["workflow_request_id"] = 1
        request_null_capability.matchdict["version_number"] = 2
        request_null_capability.matchdict["version"] = 2
        request_null_capability.matchdict["capability_name"] = fake_capability.name
        request_null_capability.matchdict["followon_type"] = "std_cms_imaging"
        request_null_capability.workflow_service = fake_wf_service
        request_null_capability.json_body = {
            "versions": [fake_v1.__json__(), fake_v2.__json__()],
        }
        fake_file_content = '{"destinationDirectory": "/var/tmp", "fileSetIds": "[]"}'

        with patch(
            "workspaces.workflow.services.workflow_service.WorkflowService.retrieve_file_content",
            return_value=fake_file_content,
        ):
            with patch(
                "workspaces.capability.schema.CapabilityRequest.__json__",
                return_value=request_null_capability.json_body,
            ):
                response = create_follow_on_capability_request(request_null_capability)
                assert response.status_code == http.HTTPStatus.OK

                response_json = response.json_body
                assert len(response_json["versions"]) == 2
                v2_json = response_json["versions"][1]
                exec_json = v2_json["current_execution"]
                assert exec_json["state_name"] == CapabilityRequestState.Complete.name
    finally:
        request_null_capability.matchdict = original_matchdict


def test_create_version_from_previous_execution_script(request_null_capability: DummyRequest):
    """
    Test for capability_execution.create_version_from_previous_execution_script():
    is the route configured as expected?
    should be capability/request/execution/{wf_request_id}/create_version

    :param request_null_capability: a real CapabilityRequest in alternate universe
    :return:
    """

    # N.B. doing it this way because using `with pytest.raises(...)`, we never make it into the `with` block
    # even though we should; thanks, Python
    try:
        create_version_from_previous_execution_script(request_null_capability)
    except KeyError as a_err:
        assert f"{a_err}" == "'wf_request_id'"


def test_assign_capability_request(request_null_capability: DummyRequest):
    """
    Tests that a specified DA/AOD is assigned to a capability request

    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    :return:
    """
    original_matchdict = request_null_capability.matchdict
    try:
        request_id = request_null_capability.matchdict["request_id"] = 1

        request_null_capability.matchdict["user_name"] = "debbie_da"
        request_null_capability.matchdict["group"] = "DA"
        response = assign_capability_request(request_null_capability)
        assert response.status_code == http.HTTPStatus.OK

        request_null_capability.matchdict["user_name"] = "alice_aod"
        request_null_capability.matchdict["group"] = "AOD"
        response = assign_capability_request(request_null_capability)
        assert response.status_code == http.HTTPStatus.OK

        # confirm that DA and AOD have been assigned
        capability_request = request_null_capability.capability_info.lookup_capability_request(request_id)
        assert capability_request.assigned_da == "debbie_da"
        assert capability_request.assigned_aod == "alice_aod"

    finally:
        request_null_capability.matchdict = original_matchdict


def test_get_available_staff_by_group(request_null_capability: DummyRequest):
    """
    Test for get_available_qa_staff_in_group() endpoint

    :param request_null_capability: wants to be a CapabilityRequest when it grows up
    :return:
    """
    original_matchdict = request_null_capability.matchdict

    da_group = "DA"
    request_null_capability.matchdict["group"] = da_group

    try:
        avail_das = request_null_capability.capability_info.get_available_staff_by_group(da_group)
        for da in avail_das:
            assert da.group == da_group
            assert da.available
    finally:
        request_null_capability.matchdict = original_matchdict
